/* ====================================================================
 *  Copyright 2003-2004 Paul Querna
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */

/**
 * mod_vhost_dbi - Mass Virtual Hosting by storing Server Information in
 * a libdbi supported database.
 */

#include "apr.h"
#include "apr_strings.h"
#include "apr_lib.h"
#include "apr_uri.h"

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_request.h"

/* Our DBI Pooling Module */
#include "mod_dbi_pool.h"

/* libdbi Headers */
#include <dbi/dbi.h>

#if !defined(WIN32) && !defined(OS2) && !defined(BEOS) && !defined(NETWARE)
#define HAVE_UNIX_SUEXEC
#endif

#ifdef HAVE_UNIX_SUEXEC
#include "unixd.h"              /* Contains the suexec_identity hook used on Unix */
#endif

/**
 * Include php support. - not enabled. we are going for Suexec Instead of
 * Playing with the native PHP OpenBaseDir Stuff.
 */
/* #define HAVE_MOD_PHP_SUPPORT */

#ifdef HAVE_MOD_PHP_SUPPORT
#include <stdlib.h>
#include "zend.h"
#include "zend_qsort.h"
#include "zend_API.h"
#include "zend_ini.h"
#include "zend_alloc.h"
#include "zend_operators.h"
#endif


AP_DECLARE_DATA module vhost_dbi_module;

typedef struct
{
    /* Is Vhost Dbi on or off? */
    int status;
    /* String Identifier of our DBI Connection Pool */
    conn_id id;
    /* SQL Query to use for finding the Virtual Host Information */
    const char *query;
} vhost_dbi_config_rec;

#define VDBI_OFF 0
#define VDBI_ON 1

/**
 * Creates a Vhost Dbi Server Config Structure
 *
 * @param p    Pool to use for Allocations
 * @param s    Server Record to use.
 * @return     The Allocated Vhost Dbi Config Structure
 */
static void *vhost_dbi_create_server_config(apr_pool_t * p, server_rec * s)
{
    vhost_dbi_config_rec *vdbir =
        (vhost_dbi_config_rec *) apr_pcalloc(p, sizeof(vhost_dbi_config_rec));

    vdbir->status = VDBI_OFF;
    vdbir->id = NULL;
    vdbir->query = NULL;
    return vdbir;
}

#define HOSTNAME_MAGIC_CHAR "."

/**
 * Hostname Expansion Function:  
 *  "foo.home.com" -> "'foo.home.com','*.foo.home.com',
 *   '*.home.com','*.com','*'"
 * 
 * @param pool      Pool to use for Allocations
 * @param hostname  Hostname to expand
 * @return  Expanded Hostname
 */
const char *expand_hostname(apr_pool_t * pool, const char *hostname)
{
    char *host;
    char *s;
    char *state;
    char *tmp = "";
    host = apr_pstrdup(pool, hostname);

    if (!(s = apr_strtok(host, HOSTNAME_MAGIC_CHAR, &state))) {
        return apr_pstrdup(pool, hostname);
    }

    do {
        if (strlen(state) == 0) {
            tmp = apr_pstrcat(pool, tmp, "'*.", s, "', '*'", NULL);
        }
        else {
            tmp =
                apr_pstrcat(pool, tmp, "'", s, ".", state, "', ", "'*.", s,
                            ".", state, "', ", NULL);
        }
    } while ((s = apr_strtok(NULL, HOSTNAME_MAGIC_CHAR, &state)));

    return tmp;
}

#define QUERYSTRING_MAGIC_CHAR '&'
#define QUERYSTRING_LEFT_DELIM_CHAR '{'
#define QUERYSTRING_RIGHT_DELIM_CHAR '}'
#define EMPTY_VAR ""            /* do NOT set this to NULL! */

/**
 * Expand an SQL Query String with all user requested variables for 
 * the current Request 
 *
 * @param r    Request Record for the Active Request
 * @param dbi  libdbi Handle to use
 * @param querystring SQL Query to expand
 * @param host Hostname for the current request
 * @return     The Expanded SQL Query
 */
const char *populate_querystring(request_rec * r, dbi_conn * dbi,
                                 const char *querystring, const char *host)
{
    char tmp[MAX_STRING_LEN];   /* 8 KByte should be enough for everyone :) */
    const char *s, *e;
    char *p;

    tmp[0] = '\0';
    int written = 0;

    if (!(s = ap_strchr_c(querystring, QUERYSTRING_MAGIC_CHAR)))
        return querystring;

    do {
        written += (s - querystring);
        /* MAX_STRING_LEN-1 so we still have room for a null at the end */
        if (written >= MAX_STRING_LEN - 1) {
            ap_log_perror(APLOG_MARK, APLOG_ERR, 0, r->pool,
                          "[mod_vhost_dbi] Populated string would exceed %d bytes",
                          MAX_STRING_LEN);
            return NULL;
        }
        strncat(tmp, querystring, s - querystring);

        if ((s[1] == QUERYSTRING_LEFT_DELIM_CHAR)
            && (e = ap_strchr_c(s, QUERYSTRING_RIGHT_DELIM_CHAR))) {
            const char *e2 = e;
            char *var;

            p = NULL;
            querystring = e + 1;
            e = NULL;
            var = apr_pstrndup(r->pool, s + 2, e2 - (s + 2));

            if (!strcasecmp(var, "RequestHostname")) {
                e = (r->hostname ? r->hostname : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "RemoteIP")) {
                e = (r->connection->remote_ip ? r->connection->
                     remote_ip : EMPTY_VAR);
            }

            /*
             * Everything but the variable values representing fieldnames and tables gets
             * * escaped according to the selected driver 
             */

            if (e != NULL) {
                p = strdup(e);
                dbi_driver_quote_string(dbi_conn_get_driver(dbi), &p);
            }

            if (!strcasecmp(var, "ExpandHostname")) {
                e = (r->
                     hostname ? expand_hostname(r->pool, host) : EMPTY_VAR);
            }

            if (e == NULL) {
                ap_log_perror(APLOG_MARK, APLOG_WARNING, 0, r->pool,
                              "[mod_vhost_dbi] Unknown variable: %s", var);
                return NULL;
            }

            if (p == NULL) {
                p = strdup(e);
            }
            written += strlen(p);
            if (written >= MAX_STRING_LEN - 1) {
                ap_log_perror(APLOG_MARK, APLOG_ERR, 0, r->pool,
                              "[mod_vhost_dbi] Populated string would exceed %d bytes",
                              MAX_STRING_LEN);
                free(p);
                return NULL;
            }
            strcat(tmp, p);
            free(p);

        }
        else {
            ap_log_perror(APLOG_MARK, APLOG_WARNING, 0, r->pool,
                          "[mod_vhost_dbi] Invalid querystring");
            return NULL;

        };

    } while ((s = ap_strchr_c(querystring, QUERYSTRING_MAGIC_CHAR)));

    strcat(tmp, querystring);
    written += strlen(querystring);

    return apr_pstrdup(r->pool, tmp);
}

/**
 * Multi-use Configuration Switch
 *
 * @param cmd    Params for this Command
 * @param config  A Void* to our config struct
 * @param value SQL Query to expand
 * @return String to describe an error, NULL on no errors
 */
static const char *set_switch(cmd_parms * cmd, void *config,
                              const char *value)
{
    vhost_dbi_config_rec *vdbir =
        (vhost_dbi_config_rec *) ap_get_module_config(cmd->server->
                                                      module_config,
                                                      &vhost_dbi_module);

    apr_ssize_t pos = (apr_ssize_t) cmd->info;
    switch (pos) {
    case 0:                    /* VhostDbiConnName */
        vdbir->id = value;
        break;
    case 1:                    /* VhostDbiQuery */
        vdbir->query = value;
        break;
    case 2:                    /* VhostDbiEnabled */
        if (!strcasecmp(value, "On")) {
            vdbir->status = VDBI_ON;
        }
        else {
            vdbir->status = VDBI_OFF;
        }
        break;
    default:
        break;
    }
    return NULL;
}

static const command_rec vhost_dbi_commands[] = {
    AP_INIT_TAKE1("VhostDbiConnName",
                  set_switch,
                  (void *) 0,
                  RSRC_CONF,
                  "name of the dbi_pool connection to use"),
    AP_INIT_TAKE1("VhostDbiQuery",
                  set_switch,
                  (void *) 1,
                  RSRC_CONF,
                  "SQL Query to Execute to find the Domain Info"),
    AP_INIT_TAKE1("VhostDbiEnabled",
                  set_switch,
                  (void *) 2,
                  RSRC_CONF,
                  "turn mod_vhost_dbi on or off."),
    {NULL}
};

/**
 * Find the Path for the DocumentRoot based on the current request
 *
 * @param r     Request Record for the Active Request
 * @param host  Hostname for the current request
 * @param env   Enviroment to Pass on
 * @param vdbir Configuration Structure
 * @return      The DocumentRoot for this Request
 */
static char *find_path(request_rec * r, const char *host, char **env,
                       vhost_dbi_config_rec * vdbir)
{
    dbi_result res;
    dbi_conn *dbi = NULL;
    const char *qstr;
    char *path;
    char *user;
    int dbi_err;
    const char *dbi_err_str;
    dbi = dbipool_open(vdbir->id);

    if (dbi == NULL) {
        /* TODO: Shouldn't this be a really fatal error? */
        ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, r->server,
                     "vhost_dbi: find_path: DBI connection was not acquired!");
        return NULL;
    }

    qstr = populate_querystring(r, dbi, vdbir->query, r->hostname);

    res = dbi_conn_query(dbi, qstr);

    if (res == NULL) {
        /**
         * Our Query Just Failed. Ouch. 
         * Lets log it so someone can look at it later. 
         */
        dbi_err = dbi_conn_error(dbi, &dbi_err_str);
        ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, r->server,
                     "vhost_dbi:  get_path: SQL Error: (%d) %s on query: %s",
                     dbi_err, dbi_err_str, qstr);
        dbi_result_free(res);
        dbipool_close(vdbir->id, dbi);
        return NULL;
    }

    switch (dbi_result_get_numrows(res)) {
    case 1:
        break;
    case 0:
        ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, r->server,
                     "vhost_dbi: get_path: no results for %s", host);
        goto bad_res;
    default:
        ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, r->server,
                     "vhost_dbi: get_path: %s has more than 1 server row, failing.",
                     host);
        goto bad_res;
    }

    if (dbi_result_next_row(res) != 0) {
        r->hostname = apr_pstrdup(r->pool,
                                  dbi_result_get_string(res, "ServerName"));
        path = apr_pstrdup(r->pool,
                           dbi_result_get_string(res, "DocumentRoot"));
        user = apr_pstrdup(r->pool, dbi_result_get_string(res, "User"));

        apr_table_setn(r->notes, "mod_vhost_dbi_user", user);

        dbi_result_free(res);
        dbipool_close(vdbir->id, dbi);

        /* TODO: Disable for Production Servers */
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, r->server,
                     "get_path: %s from SQL: %s", path, qstr);

        return path;
    }
    else {
      bad_res:
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, r->server,
                     "No Results on query: %s", qstr);
        dbi_result_free(res);
        dbipool_close(vdbir->id, dbi);
        return NULL;
    }
}

/**
 * Translate a Host into a new Path using our DBI Backend.
 *
 * @param r     Request Record for the Active Request
 * @return      Results of our handler.
 */
static int vhost_dbi_translate_name(request_rec * r)
{
    vhost_dbi_config_rec *vdbir =
        (vhost_dbi_config_rec *) ap_get_module_config(r->server->
                                                      module_config,
                                                      &vhost_dbi_module);
    char *path;
    char *env = NULL;

    if (vdbir->status == VDBI_OFF) {
        /* We Are Not Configured to Handle this Virtual Host, 
           let it procede like a normal request... */
        return DECLINED;
    }

    /**
     *  TODO: We assume r->hostname is safe -- this isn't always true:
     *        See fix_hostname(request_rec *r) in sever/vhost.c for details
     *        At the same time, there are hundreds of other places where
     *        Apache does assume it is safe.... So if there are ever problems, 
     *        they should be fixed in the fix_hostname() function.
     *
     *        This is a mess was brought to you by International Domain Names... 
     *          VeriSign: Doing thier best to screw everyone...
     */

    path = find_path(r, r->hostname, &env, vdbir);

    if (!path) {
        ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server,
                     "vhost_dbi_translate_name: no host found in database for '%s', using defaults.",
                     r->hostname);
        return DECLINED;
    }

    apr_table_set(r->subprocess_env, "vhost_dbi_PATH", path);

    r->filename = apr_psprintf(r->pool, "%s%s", path, r->uri);

#ifdef HAVE_MOD_PHP_SUPPORT
    zend_alter_ini_entry("safe_mode", 13, "on", strlen("on"), 4, 16);
    zend_alter_ini_entry("open_basedir", 13, path, strlen(path), 4, 16);
#endif

    return OK;
}

#ifdef HAVE_UNIX_SUEXEC
static ap_unix_identity_t *get_suexec_id_doer(const request_rec * r)
{
    ap_unix_identity_t *ugid = NULL;

    const char *username = apr_table_get(r->notes, "mod_vhost_dbi_user");

    if (username == NULL) {
        return NULL;
    }

    if ((ugid = apr_palloc(r->pool, sizeof(ap_unix_identity_t *))) == NULL) {
        return NULL;
    }

    if (apr_uid_get(&ugid->uid, &ugid->gid, username, r->pool) != APR_SUCCESS) {
        return NULL;
    }

    ugid->userdir = 0;

    return ugid;
}
#endif

static int vhost_dbi_init_handler(apr_pool_t * pconf, apr_pool_t * plog,
                                  apr_pool_t * ptemp, server_rec * s)
{

    ap_add_version_component(pconf, "mod_vhost_dbi/0.1.0");

    return OK;
}

/**
 * Register our Hooks within Apache
 */
static void register_hooks(apr_pool_t * p)
{
    /* Require Some Modules to be loaded first */
    static const char *const aszPre[] =
        { "mod_alias.c", "mod_userdir.c", NULL };
    ap_hook_post_config(vhost_dbi_init_handler, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_translate_name(vhost_dbi_translate_name, aszPre, NULL,
                           APR_HOOK_FIRST);
#ifdef HAVE_UNIX_SUEXEC
    ap_hook_get_suexec_identity(get_suexec_id_doer, NULL, NULL,
                                APR_HOOK_FIRST + 1);
#endif
}


AP_DECLARE_DATA module vhost_dbi_module = {
    STANDARD20_MODULE_STUFF,
    NULL,                       /* create per-directory config structure */
    NULL,                       /* merge per-directory config structures */
    vhost_dbi_create_server_config,     /* create per-server config structure */
    NULL,                       /* merge per-server config structures */
    vhost_dbi_commands,         /* command apr_table_t */
    register_hooks              /* register hooks */
};
